
Deck = Core.class(Sprite)

function Deck:init()
  -- load our texturePack
--  self.pack = TexturePack.new("images/cardsets/default.txt", "images/cardsets/default.png")
  self:reset()
  self:loadCardSet("default")

  self:addEventListener(Event.TOUCHES_BEGIN, self.onTouch, self)
  self:addEventListener(Event.TOUCHES_MOVE, self.onDrag, self)
  self:addEventListener(Event.TOUCHES_END, self.onRelease, self)
end

-- set new mode
function Deck:setMode(mode)
  self.selectionMode = mode
end

-- reset for a new game
function Deck:reset()
  -- remove any previously attached cards
  local n = self:getNumChildren()
  if n > 0 then
    for i = n, 1, -1 do
      self:removeChildAt(i)
    end
  end
  
  self.cards = {} -- dealt cards is now empty
  
  self.suites = { -- suites of cards that can be dealt
    {id = "clubs", remaining = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13}},
    {id = "spades", remaining = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13}},
    {id = "diamonds", remaining = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13}},
    {id = "hearts", remaining = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13}},
  }


  
  self.cardsLeft = 52 -- number of cards left to deal
  
  self.selectionMode = "normal"
  self.selected = nil
  self.target = nil
end

-- return a text string with the current state of the class
function Deck:countCards()
  return "in hand=" .. 52 - self.cardsLeft .. ", deck=" .. self.cardsLeft
end

-- returns the table of currently dealt cards
function Deck:getCards()
  return self.cards
end

-- load a new set of card graphics
function Deck:loadCardSet(packname)
  local pack = TexturePack.new("images/cardsets/" .. packname .. ".txt", "images/cardsets/" .. packname .. ".png")
  self.pack = pack

  -- recalculate spread data
  local region = self.pack:getTextureRegion("back.png")
  local x, y, w, h = region:getRegion()
  
  self.spreadData = {
    x = application:getContentWidth() * 0.5, -- center of spread
    y = application:getContentHeight() * 0.85,
    d = h * 1.5, -- distance from center of spread
    s = 90,
  }

  -- redraw all visible cards using new graphics
  local cards = self.cards
  if #cards > 0 then
    for i = 1, #cards do
      local card = cards[i]
      if card.faceUp then
        card:setTextureRegion(pack:getTextureRegion(card.suite .. "/" .. card.value .. ".png"))
      else
        card:setTextureRegion(pack:getTextureRegion("back.png"))
      end
    end
  end
end

-- unselect all selected cards
function Deck:unselectAll()
  local cards = self.cards
  for i = 1, #cards do
    local card = cards[i]
    card:setColorTransform(1, 1, 1)
    card.discard = false
  end
  self.target = nil
  self.selected = nil
end

-- unselect all selected cards
function Deck:discard()
  local remove = table.remove
  local cards = self.cards
  for i = #cards, 1, -1 do
    local card = cards[i]
    if card.discard then
      remove(cards, i)
      self:removeChildAt(i)
    end
  end
  self.target = nil
  self.selected = nil
end

-- pick a card.. any card..
function Deck:dealCard()
  
  -- exit if mode is not normal
  if self.selectionMode ~= "normal" then return end
  
  local random, remove = math.random, table.remove
  local pack = self.pack
  local cards = self.cards
  
  if self.cardsLeft > 0 then
    self.cardsLeft -= 1
    
    local suites = self.suites
    local availableSuites = {}
    for i = 1, #suites do
      if #suites[i].remaining > 0 then
        availableSuites[#availableSuites + 1] = i
      end
    end
    
    -- get suite to deal a card from
    local d = random(1, #availableSuites)
    local suite = self.suites[availableSuites[d]] -- the suite we will select from
    
    -- get a random card value from the suite
    local remaining = suite.remaining
    local value = remove(remaining, random(1, #remaining))
    
    -- create the new card
    local card = Card.new(pack:getTextureRegion("back.png"),{
      suite = suite.id,
      value = value,
    })
    card:setTextureRegion(pack:getTextureRegion(suite.id .. "/" .. value .. ".png"))
    
    cards[#cards + 1] = card
    self:addChild(card)
    self:spreadHand()
    
    return card
  end
  
  return false
end

-- fan that deck baby
function Deck:spreadHand()
  local sin, cos = math.sin, math.cos
  local cards = self.cards
  
  local p = self.spreadData
  local ox = p.x -- center of spread
  local oy = p.y
  local d = p.d -- distance from center or spread
  
  local s = p.s -- angle of spread
  local a = -s * 0.5 % 360
  
  for i = 1, #cards do
    local card = cards[i]
    
    -- rotate and position card
    local x = ox + d * cos(^<(a - 90))
    local y = oy + d * sin(^<(a - 90))
    card:setCardRotation(a)
    card:setCardPosition(x, y)
    
    a += (s / #cards) -- angle of next card in hand
  end
end

-- determine which card was touched (if any)
function Deck:getTouched(tx, ty)
  local cards = self.cards
  local left, right, leftTouched, rightTouched, touched
  
  if #cards == 1 then
    
    -- special case for only one card in hand
    left = cards[1]
    if cards[1]:containsPoint(tx, ty) then
      touched = left
    end
  
  else
  
    -- there are more than 1 card in hand
    for i = 1, #cards - 1 do
      left = cards[i]
      right = cards[i + 1]
      leftTouched = left:containsPoint(tx, ty) 
      rightTouched = right:containsPoint(tx, ty) 
      
      -- if the left was touched but not the right then this is the selected card
      if leftTouched and not rightTouched then
        touched = left
      end
    end
  end
  
  -- special case if last card was touched (and there are multiple cards in hand)
  if rightTouched then
    touched = right
  end
  
  return touched
end


-- determine which card was touched (if any)
function Deck:onTouch(e)
  -- where the screen was touched
  local tx, ty = e.touch.x, e.touch.y
  local touched = self:getTouched(tx, ty)
  
  if touched then

    if self.selectionMode ~= "normal" then
    
      -- toggle discard state for selected card
      if touched then
        if touched.discard then
          touched:setColorTransform(1, 1, 1)
          touched.discard = false
        else
          touched:setColorTransform(1, 0.5, 0.5)
          touched.discard = true
        end
      end
    
    else
    
      -- unselect last selected card (if any)
      if self.selected then
        self.selected:setColorTransform(1, 1, 1)
      end
      
      -- highlight currently selected card
      if touched then
        touched:setColorTransform(0.5, 1, 0.5)
        self.selected = touched
      end
    
    
    end
  
  else
    
    -- no card was clicked so unselect last selected card (if any)
    if self.selected then
      self.selected:setColorTransform(1, 1, 1)
    end

  end
end

-- track target card for drag'n'drop
function Deck:onDrag(e)
  
  -- exit if mode is not normal
  if self.selectionMode ~= "normal" then return end
  
  -- only track a target if there is something selected
  local selected = self.selected
  if selected then
    local tx, ty = e.touch.x, e.touch.y
    local target = self:getTouched(tx, ty)
    
    if target then
      if target ~= selected then
        -- unselect last selected target (if any)
        if self.target then
          self.target:setColorTransform(1, 1, 1)
        end
        -- highlight currently selected target
        target:setColorTransform(0.5, 0.5, 1)
        self.target = target
      end
    end
  end
end

-- swap cards of selected dropped onto a different card
function Deck:onRelease(e)
  
  -- exit if mode is not normal
  if self.selectionMode ~= "normal" then return end
  
  local selected, target = self.selected, self.target
  
  if selected and target then
    if target ~= selected then
      local pack = self.pack
      
      -- unselect both
      selected:setColorTransform(1, 1, 1)
      target:setColorTransform(1, 1, 1)
      
      -- get info
      local ss, sv = selected.suite, selected.value
      local ts, tv = target.suite, target.value
      
      -- exchange info
      target.suite, target.value = ss, sv
      selected.suite, selected.value = ts, tv
      
      -- update graphics
      target:setTextureRegion(pack:getTextureRegion(target.suite .. "/" .. target.value .. ".png"))
      selected:setTextureRegion(pack:getTextureRegion(selected.suite .. "/" .. selected.value .. ".png"))
      
      -- clear for next time
      self.selected = nil
      self.target = nil
    end
  end
end
